package ch.psi.daq.queryrest.query;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.commons.lang3.tuple.Triple;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.collect.Streams;

import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.common.tuple.Quadruple;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.PulseIdTime;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.backend.BackendAccess;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQConfigQuery;
import ch.psi.daq.domain.query.DAQConfigQueryElement;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.RangeQuery;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.backend.BackendQueryImpl;
import ch.psi.daq.domain.query.backend.analyzer.BackendQueryAnalyzer;
import ch.psi.daq.domain.query.channels.BackendsChannelConfigurationCache;
import ch.psi.daq.domain.query.channels.ChannelConfigurationsRequest;
import ch.psi.daq.domain.query.channels.ChannelConfigurationsResponse;
import ch.psi.daq.domain.query.channels.ChannelsRequest;
import ch.psi.daq.domain.query.channels.ChannelsResponse;
import ch.psi.daq.domain.query.channels.LongHash;
import ch.psi.daq.domain.query.channels.RangeQueryResponse;
import ch.psi.daq.domain.query.operation.EventField;
import ch.psi.daq.domain.query.processor.QueryProcessor;
import ch.psi.daq.domain.reader.RequestRangeQueryResult;

public abstract class AbstractQueryManager implements QueryManager {
   private static final Logger LOGGER = LoggerFactory.getLogger(AbstractQueryManager.class);

   private BackendsChannelConfigurationCache channelsCache;
   private Function<BackendQuery, BackendQueryAnalyzer> queryAnalizerFactory;

   protected void init(
         final BackendsChannelConfigurationCache channelsCache,
         final Function<BackendQuery, BackendQueryAnalyzer> queryAnalizerFactory) {
      this.channelsCache = channelsCache;
      this.queryAnalizerFactory = queryAnalizerFactory;
   }

   protected BackendsChannelConfigurationCache getConfigurationCache() {
      return channelsCache;
   }

   protected Function<BackendQuery, BackendQueryAnalyzer> getQueryAnalizerFactory() {
      return queryAnalizerFactory;
   }

   @Override
   public Stream<Backend> getBackends() {
      return Backend.getBackends().stream()
            .filter(backend -> getConfigurationCache().getActiveBackends().contains(backend));
   }

   @Override
   public LongHash getChannelsHash() {
      return getConfigurationCache().getChannelsHash();
   }

   @Override
   public Stream<ChannelsResponse> getChannels(
         final ChannelsRequest request) {
      return getConfigurationCache().getChannels(request);
   }

   @Override
   public LongHash getChannelConfigurationsHash() {
      return getConfigurationCache().getChannelConfigurationsHash();
   }

   @Override
   public Stream<ChannelConfigurationsResponse> getChannelConfigurations(
         final ChannelConfigurationsRequest request) {
      return getConfigurationCache().getChannelConfigurations(request);
   }

   @Override
   public ChannelConfiguration getChannelConfiguration(
         final ChannelName channel) {
      return getConfigurationCache().getChannelConfiguration(channel);
   }

   @Override
   public Entry<DAQConfigQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>> queryConfigs(
         final DAQConfigQuery daqQuery) {
      // set backends if not defined yet
      getConfigurationCache().configureBackends(daqQuery.getChannels());

      Stream<Triple<BackendQuery, ChannelName, ?>> resultStreams =
            BackendQueryImpl
                  .getBackendQueries(daqQuery)
                  .stream()
                  .filter(
                        query -> query.getBackend().getBackendAccess().hasDataReader())
                  .flatMap(
                        query -> {
                           /* all the magic happens here */
                           final Map<String, Stream<? extends ChannelConfiguration>> channelToConfig =
                                 query.getChannelConfigurations();

                           return channelToConfig.entrySet().stream()
                                 .map(entry -> {
                                    return Triple.of(
                                          query,
                                          new ChannelName(entry.getKey(), query.getBackend()),
                                          entry.getValue());
                                 });
                        });

      return Pair.of(daqQuery, resultStreams);
   }

   @Override
   public List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>> queryEvents(
         final DAQQueries queries) {
      // set backends if not defined yet
      for (DAQQueryElement daqQuery : queries) {
         getConfigurationCache().configureBackends(daqQuery.getChannels());
      }

      final List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>> results =
            new ArrayList<>(queries.getQueries().size());

      for (final DAQQueryElement queryElement : queries) {
         Stream<Quadruple<BackendQuery, ChannelName, ?, ?>> resultStreams =
               BackendQueryImpl
                     .getBackendQueries(queryElement)
                     .stream()
                     .filter(
                           query -> {
                              return query.getBackend().getBackendAccess().hasDataReader()
                                    && query.getBackend().getBackendAccess().hasQueryProcessor();
                           })
                     .flatMap(
                           query -> {
                              final QueryProcessor processor =
                                    query.getBackend().getBackendAccess().getQueryProcessor();
                              final BackendQueryAnalyzer queryAnalizer = queryAnalizerFactory.apply(query);

                              // ChannelEvent query
                              /* all the magic happens here */
                              final Stream<Entry<ChannelName, Stream<? extends DataEvent>>> channelToDataEvents =
                                    processor.process(queryAnalizer);
                              /* do post-process */
                              final Stream<Entry<ChannelName, ?>> channelToData =
                                    queryAnalizer.postProcess(channelToDataEvents);

                              // ChannelConfig query
                              final BackendQuery configQuery =
                                    new BackendQueryImpl(query, queryElement.getConfigFields());
                              final Map<String, Stream<? extends ChannelConfiguration>> channelToConfig =
                                    configQuery.getChannelConfigurations();

                              return channelToData.map(entry -> {
                                 return Quadruple.of(
                                       query,
                                       entry.getKey(),
                                       channelToConfig.get(entry.getKey().getName()),
                                       entry.getValue());
                              });
                           });

         // Now we have a stream that loads elements sequential BackendQuery by BackendQuery.
         // By materializing the outer Stream the elements of all BackendQuery are loaded async
         // (speeds things up but requires also more memory - i.e. it relies on Backends not loading
         // all elements into memory at once)
         resultStreams = resultStreams.collect(Collectors.toList()).stream();

         results.add(Pair.of(queryElement, resultStreams));
      }

      return results;
   }

   @Override
   public Stream<RangeQueryResponse> queryRange(final RangeQuery rangeQuery) throws Exception {
      if (rangeQuery.getChannels().isEmpty()) {
         // query for the general (not channel specific) pulseId/globalTime
         final BackendAccess backendAccess = getConfigurationCache().getDefaultBackend().getBackendAccess();
         if (backendAccess.hasStreamEventReader()) {
            final CompletableFuture<RequestRangeQueryResult> future =
                  backendAccess.getStreamEventReader().getPulseIdTimeMappingAsync(rangeQuery.getRange());
            final RequestRangeQueryResult result = future.get(
                  backendAccess.getBackend().getApplicationContext()
                        .getBean(DomainConfig.BEAN_NAME_DISTRIBUTED_READ_TIMEOUT, Integer.class),
                  TimeUnit.SECONDS);
            return Stream.of(
                  new RangeQueryResponse(
                        new ChannelName(ChannelConfiguration.COMBINED_CHANNEL_NAME, backendAccess.getBackend()),
                        new PulseIdTime(result.getStart(), false),
                        new PulseIdTime(result.getEnd(), false)));
         } else {
            LOGGER.warn("Backend '{}' has no StreamEventReader. Cannot execute '{}'.", backendAccess.getBackend(),
                  rangeQuery);
            return Stream.of(
                  new RangeQueryResponse(
                        new ChannelName(ChannelConfiguration.COMBINED_CHANNEL_NAME, backendAccess.getBackend()),
                        null,
                        null));
         }
      } else {
         // set backends if not defined yet
         getConfigurationCache().configureBackends(rangeQuery.getChannels());

         final LinkedHashSet<EventField> eventFields = new LinkedHashSet<>(2);
         eventFields.add(EventField.pulseId);
         eventFields.add(EventField.globalTime);
         eventFields.add(EventField.globalDate);
         eventFields.add(EventField.globalSeconds);
         rangeQuery.setAggregation(null);
         rangeQuery.setEventFields(eventFields);
         rangeQuery.setMapping(null);
         rangeQuery.setValueTransformations(null);
         rangeQuery.setLimit(1);
         rangeQuery.setOrdering(Ordering.asc);

         Stream<Triple<ChannelName, Stream<? extends DataEvent>, Stream<? extends DataEvent>>> resultStreams =
               BackendQueryImpl
                     .getBackendQueries(rangeQuery)
                     .stream()
                     .filter(
                           query -> {
                              return query.getBackend().getBackendAccess().hasDataReader()
                                    && query.getBackend().getBackendAccess().hasQueryProcessor();
                           })
                     .flatMap(
                           query -> {
                              final BackendQueryImpl queryForward = new BackendQueryImpl(query);
                              // queryForward.setAggregation(null);
                              // queryForward.setEventFields(eventFields);
                              // queryForward.setMapping(null);
                              // queryForward.setValueTransformations(null);
                              // queryForward.setLimit(1);
                              // queryForward.setOrdering(Ordering.asc);

                              final QueryProcessor processor =
                                    queryForward.getBackend().getBackendAccess().getQueryProcessor();

                              final BackendQueryAnalyzer queryAnalizerForward =
                                    queryAnalizerFactory.apply(queryForward);
                              final Stream<Entry<ChannelName, Stream<? extends DataEvent>>> channelToDataEventForward =
                                    processor.process(queryAnalizerForward);

                              final BackendQueryImpl queryBackwards = new BackendQueryImpl(queryForward);
                              queryBackwards.setOrdering(Ordering.desc);
                              final BackendQueryAnalyzer queryAnalizerBackwards =
                                    queryAnalizerFactory.apply(queryForward);
                              final Stream<Entry<ChannelName, Stream<? extends DataEvent>>> channelToDataEventBackwards =
                                    processor.process(queryAnalizerBackwards);

                              return Streams.zip(
                                    channelToDataEventForward,
                                    channelToDataEventBackwards,
                                    (entry1, entry2) -> {
                                       return Triple.of(
                                             entry1.getKey(),
                                             entry1.getValue(),
                                             entry2.getValue());
                                       // return Triple.of(
                                       // entry1.getKey(),
                                       // Streams.concat(entry1.getValue(), entry2.getValue()));
                                    });
                           });

         // make sure queries are executed (Streams populate async)
         resultStreams = resultStreams.collect(Collectors.toList()).stream();

         // now access internal streams (which will block until data is available)
         return resultStreams
               .map(triple -> {
                  return new RangeQueryResponse(
                        triple.getLeft(),
                        new PulseIdTime(triple.getMiddle().findFirst().orElse(null), false),
                        new PulseIdTime(triple.getRight().findFirst().orElse(null), false));
               });
      }
   }
}
